from datetime import datetime

from traveltimepy import Coordinates, PublicTransport


def test_districts_departure(sdk):
    results = sdk.postcodes(
        coordinates=[Coordinates(lat=51.507609, lng=-0.128315)],
        departure_time=datetime.now(),
        transportation=PublicTransport()
    )
    assert len(results) > 0


def test_districts_arrival(sdk):
    results = sdk.postcodes(
        coordinates=[Coordinates(lat=51.507609, lng=-0.128315)],
        arrival_time=datetime.now(),
        transportation=PublicTransport()
    )
    assert len(results) > 0


def test_sectors_departure(sdk):
    results = sdk.postcodes(
        coordinates=[Coordinates(lat=51.507609, lng=-0.128315)],
        departure_time=datetime.now(),
        transportation=PublicTransport()
    )
    assert len(results) > 0


def test_sectors_arrival(sdk):
    results = sdk.postcodes(
        coordinates=[Coordinates(lat=51.507609, lng=-0.128315)],
        arrival_time=datetime.now(),
        transportation=PublicTransport()
    )
    assert len(results) > 0



from datetime import datetime
from typing import List, Optional, Union

from pydantic.main import BaseModel

from traveltimepy.dto.common import Location, Property, FullRange
from traveltimepy.dto.transportation import PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.routes import RoutesResponse
from traveltimepy.itertools import split, flatten


class ArrivalSearch(BaseModel):
    id: str
    departure_location_ids: List[str]
    arrival_location_id: str
    arrival_time: datetime
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[Property]
    range: Optional[FullRange] = None


class DepartureSearch(BaseModel):
    id: str
    arrival_location_ids: List[str]
    departure_location_id: str
    departure_time: datetime
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[Property]
    range: Optional[FullRange] = None


class RoutesRequest(TravelTimeRequest[RoutesResponse]):
    locations: List[Location]
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self) -> List[TravelTimeRequest]:
        return [
            RoutesRequest(locations=self.locations, departure_searches=departures, arrival_searches=arrivals)
            for departures, arrivals in split(self.departure_searches, self.arrival_searches, 10)
        ]

    def merge(self, responses: List[RoutesResponse]) -> RoutesResponse:
        return RoutesResponse(results=flatten([response.results for response in responses]))



import time
from datetime import datetime

from benchmarks.common import generate_coordinates
from traveltimepy import TravelTimeSdk, Driving


def generate_isochrones(size: int):
    sdk = TravelTimeSdk('APP_ID', 'API_KEY')
    coordinates = generate_coordinates(51.507609, -0.128315, 0.05, size)
    return sdk.time_map(
        coordinates=coordinates,
        transportation=Driving(),
        arrival_time=datetime.now()
    )


if __name__ == '__main__':
    start = time.perf_counter()
    response = generate_isochrones(50)
    request_time = time.perf_counter() - start
    print('Request completed in {0:.0f}s for 50 isochrones'.format(request_time))

    start = time.perf_counter()
    response2 = generate_isochrones(100)
    request_time = time.perf_counter() - start
    print('Request completed in {0:.0f}s for 100 isochrones'.format(request_time))

    start = time.perf_counter()
    response3 = generate_isochrones(300)
    request_time = time.perf_counter() - start
    print('Request completed in {0:.0f}s for 300 isochrones'.format(request_time))



from datetime import datetime
from typing import List, Optional, Union

from pydantic.main import BaseModel

from traveltimepy.dto.common import Location, FullRange, Property
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.time_filter import TimeFilterResponse
from traveltimepy.itertools import split, flatten
from traveltimepy.dto.transportation import PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain


class ArrivalSearch(BaseModel):
    id: str
    departure_location_ids: List[str]
    arrival_location_id: str
    arrival_time: datetime
    travel_time: int
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[Property]
    range: Optional[FullRange] = None


class DepartureSearch(BaseModel):
    id: str
    arrival_location_ids: List[str]
    departure_location_id: str
    departure_time: datetime
    travel_time: int
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[Property]
    range: Optional[FullRange] = None


class TimeFilterRequest(TravelTimeRequest[TimeFilterResponse]):
    locations: List[Location]
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self) -> List[TravelTimeRequest]:
        return [
            TimeFilterRequest(locations=self.locations, departure_searches=departures, arrival_searches=arrivals)
            for departures, arrivals in split(self.departure_searches, self.arrival_searches, 10)
        ]

    def merge(self, responses: List[TimeFilterResponse]) -> TimeFilterResponse:
        return TimeFilterResponse(results=flatten([response.results for response in responses]))



from datetime import datetime
from enum import Enum
from typing import List, Union, Optional

from pydantic import BaseModel

from traveltimepy.dto.common import Coordinates, FullRange
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.zones import DistrictsResponse, SectorsResponse
from traveltimepy.itertools import split, flatten
from traveltimepy.dto.transportation import PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain


class ZonesProperty(str, Enum):
    TRAVEL_TIME_REACHABLE = 'travel_time_reachable'
    TRAVEL_TIME_ALL = 'travel_time_all'
    COVERAGE = 'coverage'


class ArrivalSearch(BaseModel):
    id: str
    coords: Coordinates
    travel_time: int
    arrival_time: datetime
    reachable_postcodes_threshold: float
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[ZonesProperty]
    range: Optional[FullRange] = None


class DepartureSearch(BaseModel):
    id: str
    coords: Coordinates
    travel_time: int
    departure_time: datetime
    reachable_postcodes_threshold: float
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain]
    properties: List[ZonesProperty]
    range: Optional[FullRange] = None


class SectorsRequest(TravelTimeRequest[SectorsResponse]):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self) -> List[TravelTimeRequest]:
        return [
            SectorsRequest(departure_searches=departures, arrival_searches=arrivals)
            for departures, arrivals in split(self.departure_searches, self.arrival_searches, 10)
        ]

    def merge(self, responses: List[SectorsResponse]) -> SectorsResponse:
        return SectorsResponse(
            results=sorted(flatten([response.results for response in responses]), key=lambda res: res.search_id)
        )


class DistrictsRequest(TravelTimeRequest[DistrictsResponse]):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self) -> List[TravelTimeRequest]:
        return [
            DistrictsRequest(departure_searches=departures, arrival_searches=arrivals)
            for departures, arrivals in split(self.departure_searches, self.arrival_searches, 10)
        ]

    def merge(self, responses: List[DistrictsResponse]) -> DistrictsResponse:
        return DistrictsResponse(
            results=sorted(flatten([response.results for response in responses]), key=lambda res: res.search_id)
        )



from datetime import datetime
from typing import List, Optional, Dict, Union

from traveltimepy.dto.common import Location, Coordinates, Rectangle, Property, FullRange, Range
from traveltimepy.dto.transportation import PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain
from traveltimepy.dto.requests.zones import ZonesProperty
from traveltimepy.dto.requests.time_filter_proto import ProtoCountry, ProtoTransportation
from traveltimepy.dto.requests.time_filter_fast import Transportation

from traveltimepy.version import __version__
from traveltimepy.accept_type import AcceptType
from traveltimepy.itertools import join_opt
from traveltimepy.dto.requests.supported_locations import SupportedLocationsRequest

from traveltimepy.dto.responses.map_info import MapInfoResponse, Map
from traveltimepy.dto.responses.postcodes import PostcodesResponse, PostcodesResult
from traveltimepy.dto.responses.routes import RoutesResponse, RoutesResult
from traveltimepy.dto.responses.supported_locations import SupportedLocationsResponse
from traveltimepy.dto.responses.time_filter import TimeFilterResponse, TimeFilterResult
from traveltimepy.dto.responses.time_filter_fast import TimeFilterFastResponse, TimeFilterFastResult
from traveltimepy.dto.responses.time_map import TimeMapResponse, TimeMapResult
from traveltimepy.dto.responses.zones import DistrictsResponse, SectorsResponse, DistrictsResult, SectorsResult

from traveltimepy.mapper import (
    create_time_filter,
    create_time_filter_fast,
    create_postcodes,
    create_districts,
    create_sectors,
    create_routes,
    create_proto_request,
    create_time_map,
    create_intersection, create_union
)

from traveltimepy.proto_http import send_proto, send_proto_async
from traveltimepy.http import (
    send_get,
    send_get_async,
    send_post,
    send_post_async,
    SdkParams
)

from geojson_pydantic import FeatureCollection


class TravelTimeSdk:

    def __init__(self, app_id: str, api_key: str, limit_per_host: int = 2, host: str = 'api.traveltimeapp.com') -> None:
        self.__app_id = app_id
        self.__api_key = api_key
        self.__sdk_params = SdkParams(host, limit_per_host)

    async def time_filter_async(
        self,
        locations: List[Location],
        search_ids: Dict[str, List[str]],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        properties: Optional[List[Property]] = None,
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        range: Optional[FullRange] = None
    ) -> List[TimeFilterResult]:
        resp = await send_post_async(
            TimeFilterResponse,
            'time-filter',
            self.__headers(AcceptType.JSON),
            create_time_filter(
                locations,
                search_ids,
                transportation,
                properties,
                departure_time,
                arrival_time,
                travel_time,
                range
            ),
            self.__sdk_params
        )

        return resp.results

    def time_filter(
        self,
        locations: List[Location],
        search_ids: Dict[str, List[str]],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        properties: Optional[List[Property]] = None,
        travel_time: int = 3600,
        range: Optional[FullRange] = None
    ) -> List[TimeFilterResult]:
        return send_post(
            TimeFilterResponse,
            'time-filter',
            self.__headers(AcceptType.JSON),
            create_time_filter(
                locations,
                search_ids,
                transportation,
                properties,
                departure_time,
                arrival_time,
                travel_time,
                range
            ),
            self.__sdk_params
        ).results

    async def map_info_async(self) -> List[Map]:
        res = await send_get_async(MapInfoResponse, 'map-info', self.__headers(AcceptType.JSON), self.__sdk_params)
        return res.maps

    def map_info(self) -> List[Map]:
        return send_get(MapInfoResponse, 'map-info', self.__headers(AcceptType.JSON), self.__sdk_params).maps

    async def geocoding_async(
        self,
        query: str,
        limit: Optional[int] = None,
        within_countries: Optional[List[str]] = None,
        format_name: Optional[bool] = None,
        format_exclude_country: Optional[bool] = None,
        bounds: Optional[Rectangle] = None
    ) -> FeatureCollection:
        return await send_get_async(
            FeatureCollection,
            'geocoding/search',
            self.__headers(AcceptType.JSON),
            self.__sdk_params,
            self.__geocoding_params(
                query,
                limit,
                within_countries,
                format_name,
                format_exclude_country,
                bounds
            )
        )

    def geocoding(
        self,
        query: str,
        limit: Optional[int] = None,
        within_countries: Optional[List[str]] = None,
        format_name: Optional[bool] = None,
        format_exclude_country: Optional[bool] = None,
        bounds: Optional[Rectangle] = None
    ) -> FeatureCollection:
        return send_get(
            FeatureCollection,
            'geocoding/search',
            self.__headers(AcceptType.JSON),
            self.__sdk_params,
            self.__geocoding_params(
                query,
                limit,
                within_countries,
                format_name,
                format_exclude_country,
                bounds
            )
        )

    async def geocoding_reverse_async(
        self,
        lat: float,
        lng: float,
        within_countries: Optional[List[str]] = None
    ) -> FeatureCollection:
        return await send_get_async(
            FeatureCollection,
            'geocoding/reverse',
            self.__headers(AcceptType.JSON),
            self.__sdk_params,
            self.__geocoding_reverse_params(lat, lng, within_countries)
        )

    def geocoding_reverse(
        self,
        lat: float,
        lng: float
    ) -> FeatureCollection:
        return send_get(
            FeatureCollection,
            'geocoding/reverse',
            self.__headers(AcceptType.JSON),
            self.__sdk_params,
            self.__geocoding_reverse_params(lat, lng)
        )

    async def supported_locations_async(self, locations: List[Location]) -> SupportedLocationsResponse:
        return await send_post_async(
            SupportedLocationsResponse,
            'supported-locations',
            self.__headers(AcceptType.JSON),
            SupportedLocationsRequest(locations=locations),
            self.__sdk_params
        )

    def supported_locations(self, locations: List[Location]) -> SupportedLocationsResponse:
        return send_post(
            SupportedLocationsResponse,
            'supported-locations',
            self.__headers(AcceptType.JSON),
            SupportedLocationsRequest(locations=locations),
            self.__sdk_params
        )

    async def time_filter_fast_async(
        self,
        locations: List[Location],
        search_ids: Dict[str, List[str]],
        transportation: Transportation,
        travel_time: int = 3600,
        properties: Optional[List[Property]] = None,
        one_to_many: bool = False
    ) -> List[TimeFilterFastResult]:
        resp = await send_post_async(
            TimeFilterFastResponse,
            'time-filter/fast',
            self.__headers(AcceptType.JSON),
            create_time_filter_fast(
                locations,
                search_ids,
                transportation,
                travel_time,
                properties,
                one_to_many
            ),
            self.__sdk_params
        )
        return resp.results

    def time_filter_fast(
        self,
        locations: List[Location],
        search_ids: Dict[str, List[str]],
        transportation: Transportation,
        travel_time: int = 3600,
        properties: Optional[List[Property]] = None,
        one_to_many: bool = False
    ) -> List[TimeFilterFastResult]:
        return send_post(
            TimeFilterFastResponse,
            'time-filter/fast',
            self.__headers(AcceptType.JSON),
            create_time_filter_fast(
                locations,
                search_ids,
                transportation,
                travel_time,
                properties,
                one_to_many
            ),
            self.__sdk_params
        ).results

    async def postcodes_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 1800,
        properties: Optional[List[Property]] = None,
        range: Optional[FullRange] = None
    ) -> List[PostcodesResult]:
        resp = await send_post_async(
            PostcodesResponse,
            'time-filter/postcodes',
            self.__headers(AcceptType.JSON),
            create_postcodes(
                coordinates,
                departure_time,
                arrival_time,
                transportation,
                travel_time,
                properties,
                range
            ),
            self.__sdk_params
        )
        return resp.results

    def postcodes(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 1800,
        properties: Optional[List[Property]] = None,
        range: Optional[FullRange] = None
    ) -> List[PostcodesResult]:
        return send_post(
            PostcodesResponse,
            'time-filter/postcodes',
            self.__headers(AcceptType.JSON),
            create_postcodes(
                coordinates,
                departure_time,
                arrival_time,
                transportation,
                travel_time,
                properties,
                range
            ),
            self.__sdk_params
        ).results

    async def districts_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        travel_time: int = 1800,
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        reachable_postcodes_threshold=0.1,
        properties: Optional[List[ZonesProperty]] = None,
        range: Optional[FullRange] = None
    ) -> List[DistrictsResult]:
        res = await send_post_async(
            DistrictsResponse,
            'time-filter/postcode-districts',
            self.__headers(AcceptType.JSON),
            create_districts(
                coordinates,
                transportation,
                travel_time,
                departure_time,
                arrival_time,
                reachable_postcodes_threshold,
                properties,
                range
            ),
            self.__sdk_params
        )
        return res.results

    def districts(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        travel_time: int = 1800,
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        reachable_postcodes_threshold=0.1,
        properties: Optional[List[ZonesProperty]] = None,
        range: Optional[FullRange] = None
    ) -> List[DistrictsResult]:
        return send_post(
            DistrictsResponse,
            'time-filter/postcode-districts',
            self.__headers(AcceptType.JSON),
            create_districts(
                coordinates,
                transportation,
                travel_time,
                departure_time,
                arrival_time,
                reachable_postcodes_threshold,
                properties,
                range
            ),
            self.__sdk_params
        ).results

    async def sectors_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        travel_time: int = 1800,
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        reachable_postcodes_threshold=0.1,
        properties: Optional[List[ZonesProperty]] = None,
        range: Optional[FullRange] = None
    ) -> List[SectorsResult]:
        resp = await send_post_async(
            SectorsResponse,
            'time-filter/postcode-sectors',
            self.__headers(AcceptType.JSON),
            create_sectors(
                coordinates,
                transportation,
                travel_time,
                departure_time,
                arrival_time,
                reachable_postcodes_threshold,
                properties,
                range
            ),
            self.__sdk_params
        )
        return resp.results

    def sectors(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        travel_time: int = 1800,
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        reachable_postcodes_threshold=0.1,
        properties: Optional[List[ZonesProperty]] = None,
        range: Optional[FullRange] = None
    ) -> List[SectorsResult]:
        return send_post(
            SectorsResponse,
            'time-filter/postcode-sectors',
            self.__headers(AcceptType.JSON),
            create_sectors(
                coordinates,
                transportation,
                travel_time,
                departure_time,
                arrival_time,
                reachable_postcodes_threshold,
                properties,
                range
            ),
            self.__sdk_params
        ).results

    def routes(
        self,
        locations: List[Location],
        search_ids: Dict[str, List[str]],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        properties: Optional[List[Property]] = None,
        range: Optional[FullRange] = None
    ) -> List[RoutesResult]:
        return send_post(
            RoutesResponse,
            'routes',
            self.__headers(AcceptType.JSON),
            create_routes(
                locations,
                search_ids,
                transportation,
                departure_time,
                arrival_time,
                properties,
                range
            ),
            self.__sdk_params
        ).results

    async def routes_async(
        self,
        locations: List[Location],
        search_ids: Dict[str, List[str]],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        properties: Optional[List[Property]] = None,
        range: Optional[FullRange] = None
    ) -> List[RoutesResult]:
        resp = await send_post_async(
            RoutesResponse,
            'routes',
            self.__headers(AcceptType.JSON),
            create_routes(
                locations,
                search_ids,
                transportation,
                departure_time,
                arrival_time,
                properties,
                range
            ),
            self.__sdk_params
        )
        return resp.results

    def time_filter_proto(
        self,
        origin: Coordinates,
        destinations: List[Coordinates],
        country: ProtoCountry,
        transportation: ProtoTransportation,
        travel_time: int,
    ) -> List[int]:
        return send_proto(
            f'https://proto.api.traveltimeapp.com/api/v2/{country.value}/time-filter/fast/{transportation.value.name}',
            self.__proto_headers(),
            create_proto_request(origin, destinations, transportation, travel_time),
            self.__app_id,
            self.__api_key
        ).travel_times

    async def time_filter_proto_async(
        self,
        origin: Coordinates,
        destinations: List[Coordinates],
        country: ProtoCountry,
        transportation: ProtoTransportation,
        travel_time: int,
    ) -> List[int]:
        resp = await send_proto_async(
            f'https://proto.api.traveltimeapp.com/api/v2/{country.value}/time-filter/fast/{transportation.name}',
            self.__proto_headers(),
            create_proto_request(origin, destinations, transportation, travel_time),
            self.__app_id,
            self.__api_key
        )
        return resp.travel_times

    def time_map(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        arrival_time: Optional[datetime] = None,
        departure_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None
    ) -> List[TimeMapResult]:
        return send_post(
            TimeMapResponse,
            'time-map',
            self.__headers(AcceptType.JSON),
            create_time_map(
                coordinates,
                transportation,
                travel_time,
                arrival_time,
                departure_time,
                search_range
            ),
            self.__sdk_params
        ).results

    def intersection(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        arrival_time: Optional[datetime] = None,
        departure_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None
    ) -> TimeMapResult:
        return send_post(
            TimeMapResponse,
            'time-map',
            self.__headers(AcceptType.JSON),
            create_intersection(
                coordinates,
                transportation,
                travel_time,
                arrival_time,
                departure_time,
                search_range
            ),
            self.__sdk_params
        ).results[0]

    async def intersection_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        arrival_time: Optional[datetime] = None,
        departure_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None
    ) -> TimeMapResult:
        resp = await send_post_async(
            TimeMapResponse,
            'time-map',
            self.__headers(AcceptType.JSON),
            create_intersection(
                coordinates,
                transportation,
                travel_time,
                arrival_time,
                departure_time,
                search_range
            ),
            self.__sdk_params
        )
        return resp.results[0]

    def union(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        arrival_time: Optional[datetime] = None,
        departure_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None
    ) -> TimeMapResult:
        return send_post(
            TimeMapResponse,
            'time-map',
            self.__headers(AcceptType.JSON),
            create_union(
                coordinates,
                transportation,
                travel_time,
                arrival_time,
                departure_time,
                search_range
            ),
            self.__sdk_params
        ).results[0]

    async def union_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        arrival_time: Optional[datetime] = None,
        departure_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None
    ) -> TimeMapResult:
        resp = await send_post_async(
            TimeMapResponse,
            'time-map',
            self.__headers(AcceptType.JSON),
            create_union(
                coordinates,
                transportation,
                travel_time,
                arrival_time,
                departure_time,
                search_range
            ),
            self.__sdk_params
        )

        return resp.results[0]

    async def time_map_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
        arrival_time: Optional[datetime] = None,
        departure_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None
    ) -> List[TimeMapResult]:
        resp = await send_post_async(
            TimeMapResponse,
            'time-map',
            self.__headers(AcceptType.JSON),
            create_time_map(
                coordinates,
                transportation,
                travel_time,
                arrival_time,
                departure_time,
                search_range
            ),
            self.__sdk_params
        )
        return resp.results

    @staticmethod
    def __geocoding_reverse_params(
        lat: float,
        lng: float
    ) -> Dict[str, str]:
        full_query = {
            'lat': lat,
            'lng': lng
        }
        return {key: str(value) for (key, value) in full_query.items() if value is not None}

    @staticmethod
    def __geocoding_params(
        query: str,
        limit: Optional[int] = None,
        within_countries: Optional[List[str]] = None,
        format_name: Optional[bool] = None,
        format_exclude_country: Optional[bool] = None,
        bounds: Optional[Rectangle] = None
    ) -> Dict[str, str]:
        full_query = {
            'query': query,
            'limit': limit,
            'within.country': join_opt(within_countries, ','),
            'format.name': format_name,
            'format.exclude.country': format_exclude_country,
            'bounds': bounds.to_str() if bounds is not None else bounds
        }
        return {key: str(value) for (key, value) in full_query.items() if value is not None}

    @staticmethod
    def __proto_headers() -> Dict[str, str]:
        return {
            'Content-Type': AcceptType.OCTET_STREAM.value,
            'User-Agent': f'Travel Time Python SDK {__version__}'
        }

    def __headers(self, accept_type: AcceptType) -> Dict[str, str]:
        return {
            'X-Application-Id': self.__app_id,
            'X-Api-Key': self.__api_key,
            'User-Agent': f'Travel Time Python SDK {__version__}',
            'Content-Type': 'application/json',
            'Accept': accept_type.value
        }



import math
from datetime import datetime
from typing import Dict, Union, List, Optional

from traveltimepy.errors import ApiError
from traveltimepy import TimeFilterFastRequest_pb2

from traveltimepy.dto.common import Location, Coordinates, FullRange, Property, Range
from traveltimepy.dto.transportation import PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain
from traveltimepy.dto.requests.time_filter_fast import Transportation
from traveltimepy.dto.requests.zones import ZonesProperty

from traveltimepy.dto.requests.postcodes import PostcodesRequest
from traveltimepy.dto.requests.routes import RoutesRequest
from traveltimepy.dto.requests.time_filter import TimeFilterRequest
from traveltimepy.dto.requests.time_filter_fast import TimeFilterFastRequest
from traveltimepy.dto.requests.time_filter_proto import ProtoTransportation
from traveltimepy.dto.requests.zones import DistrictsRequest, SectorsRequest
from traveltimepy.dto.requests.time_map import TimeMapRequest


from traveltimepy.dto.requests import (
    time_filter,
    time_filter_fast,
    postcodes,
    zones,
    routes,
    time_map
)


def create_time_filter(
    locations: List[Location],
    search_ids: Dict[str, List[str]],
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
    properties: Optional[List[Property]],
    departure_time: Optional[datetime],
    arrival_time: Optional[datetime],
    travel_time: int,
    range: Optional[FullRange]
) -> TimeFilterRequest:
    if properties is None:
        properties = [Property.TRAVEL_TIME]
    if arrival_time is not None and departure_time is not None:
        raise ApiError('arrival_time and departure_time cannot be both specified')

    if arrival_time is not None:
        return TimeFilterRequest(
            locations=locations,
            arrival_searches=[
                time_filter.ArrivalSearch(
                    id=arrival_id,
                    arrival_location_id=arrival_id,
                    departure_location_ids=[departure_id for departure_id in departure_ids],
                    arrival_time=arrival_time,
                    travel_time=travel_time,
                    transportation=transportation,
                    properties=properties,
                    range=range
                )
                for arrival_id, departure_ids in search_ids.items()
            ],
            departure_searches=[]
        )
    elif departure_time is not None:
        return TimeFilterRequest(
            locations=locations,
            departure_searches=[
                time_filter.DepartureSearch(
                    id=departure_id,
                    departure_location_id=departure_id,
                    arrival_location_ids=[arrival_id for arrival_id in arrival_ids],
                    departure_time=departure_time,
                    travel_time=travel_time,
                    transportation=transportation,
                    properties=properties,
                    range=range
                )
                for departure_id, arrival_ids in search_ids.items()
            ],
            arrival_searches=[]
        )
    else:
        raise ApiError('arrival_time or departure_time should be specified')


def create_time_filter_fast(
    locations: List[Location],
    search_ids: Dict[str, List[str]],
    transportation: Transportation,
    travel_time: int = 3600,
    properties: Optional[List[Property]] = None,
    one_to_many: bool = False
) -> TimeFilterFastRequest:
    if properties is None:
        properties = [Property.TRAVEL_TIME]
    if one_to_many:
        return TimeFilterFastRequest(
            locations=locations,
            arrival_searches=time_filter_fast.ArrivalSearches(
                one_to_many=[
                    time_filter_fast.OneToMany(
                        id=departure_id,
                        departure_location_id=departure_id,
                        arrival_location_ids=arrival_ids,
                        transportation=transportation,
                        travel_time=travel_time,
                        arrival_time_period='weekday_morning',
                        properties=properties
                    )
                    for departure_id, arrival_ids in search_ids.items()
                ],
                many_to_one=[]
            )
        )
    else:
        return TimeFilterFastRequest(
            locations=locations,
            arrival_searches=time_filter_fast.ArrivalSearches(
                many_to_one=[
                    time_filter_fast.ManyToOne(
                        id=arrival_id,
                        arrival_location_id=arrival_id,
                        departure_location_ids=departure_ids,
                        transportation=transportation,
                        travel_time=travel_time,
                        arrival_time_period='weekday_morning',
                        properties=properties
                    )
                    for arrival_id, departure_ids in search_ids.items()
                ],
                one_to_many=[]
            )
        )


def create_postcodes(
    coordinates: List[Coordinates],
    departure_time: Optional[datetime],
    arrival_time: Optional[datetime],
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
    travel_time: int,
    properties: Optional[List[Property]],
    range: Optional[FullRange]
) -> PostcodesRequest:
    if properties is None:
        properties = [Property.TRAVEL_TIME]

    if arrival_time is not None and departure_time is not None:
        raise ApiError('arrival_time and departure_time cannot be both specified')

    if departure_time is not None:
        return PostcodesRequest(
            departure_searches=[
                postcodes.DepartureSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=departure_time,
                    transportation=transportation,
                    properties=properties,
                    range=range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[]
        )
    elif arrival_time is not None:
        return PostcodesRequest(
            arrival_searches=[
                postcodes.ArrivalSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=arrival_time,
                    transportation=transportation,
                    properties=properties,
                    range=range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[]
        )
    else:
        raise ApiError('arrival_time or departure_time should be specified')


def create_districts(
    coordinates: List[Coordinates],
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
    travel_time: int,
    departure_time: Optional[datetime],
    arrival_time: Optional[datetime],
    reachable_postcodes_threshold,
    properties: Optional[List[ZonesProperty]],
    range: Optional[FullRange]
) -> DistrictsRequest:
    if properties is None:
        properties = ZonesProperty.TRAVEL_TIME_ALL

    if arrival_time is not None and departure_time is not None:
        raise ApiError('arrival_time and departure_time cannot be both specified')

    if arrival_time is not None:
        return DistrictsRequest(
            arrival_searches=[
                zones.ArrivalSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=arrival_time,
                    reachable_postcodes_threshold=reachable_postcodes_threshold,
                    transportation=transportation,
                    properties=properties,
                    range=range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[]
        )
    elif departure_time is not None:
        return DistrictsRequest(
            departure_searches=[
                zones.DepartureSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=departure_time,
                    reachable_postcodes_threshold=reachable_postcodes_threshold,
                    transportation=transportation,
                    properties=properties,
                    range=range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[]
        )
    else:
        raise ApiError('arrival_time or departure_time should be specified')


def create_sectors(
    coordinates: List[Coordinates],
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
    travel_time: int,
    departure_time: Optional[datetime],
    arrival_time: Optional[datetime],
    reachable_postcodes_threshold,
    properties: Optional[List[ZonesProperty]],
    range: Optional[FullRange]
) -> SectorsRequest:
    if properties is None:
        properties = ZonesProperty.TRAVEL_TIME_ALL

    if arrival_time is not None and departure_time is not None:
        raise ApiError('arrival_time and departure_time cannot be both specified')

    if arrival_time is not None:
        return SectorsRequest(
            arrival_searches=[
                zones.ArrivalSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=arrival_time,
                    reachable_postcodes_threshold=reachable_postcodes_threshold,
                    transportation=transportation,
                    properties=properties,
                    range=range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[]
        )
    elif departure_time is not None:
        return SectorsRequest(
            departure_searches=[
                zones.DepartureSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=departure_time,
                    reachable_postcodes_threshold=reachable_postcodes_threshold,
                    transportation=transportation,
                    properties=properties,
                    range=range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[]
        )
    else:
        raise ApiError('arrival_time or departure_time should be specified')


def create_time_map(
    coordinates: List[Coordinates],
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
    travel_time: int,
    departure_time: Optional[datetime],
    arrival_time: Optional[datetime],
    search_range: Optional[Range]
) -> TimeMapRequest:
    if arrival_time is not None and departure_time is not None:
        raise ApiError('arrival_time and departure_time cannot be both specified')

    if arrival_time is not None:
        return TimeMapRequest(
            arrival_searches=[
                time_map.ArrivalSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=arrival_time,
                    transportation=transportation,
                    range=search_range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
            unions=[],
            intersections=[]
        )
    elif departure_time is not None:
        return TimeMapRequest(
            departure_searches=[
                time_map.DepartureSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=departure_time,
                    transportation=transportation,
                    range=search_range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
            unions=[],
            intersections=[]
        )
    else:
        raise ApiError('arrival_time or departure_time should be specified')


def create_intersection(
    coordinates: List[Coordinates],
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
    travel_time: int,
    departure_time: Optional[datetime],
    arrival_time: Optional[datetime],
    search_range: Optional[Range]
) -> TimeMapRequest:
    if arrival_time is not None and departure_time is not None:
        raise ApiError('arrival_time and departure_time cannot be both specified')

    if arrival_time is not None:
        return TimeMapRequest(
            arrival_searches=[
                time_map.ArrivalSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=arrival_time,
                    transportation=transportation,
                    range=search_range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
            unions=[],
            intersections=[
                time_map.Intersection(
                    id='Intersection search',
                    search_ids=[f'Search {ind}' for ind, _ in enumerate(coordinates)]
                )
            ]
        )
    elif departure_time is not None:
        return TimeMapRequest(
            departure_searches=[
                time_map.DepartureSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=departure_time,
                    transportation=transportation,
                    range=search_range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
            unions=[],
            intersections=[
                time_map.Intersection(
                    id='Intersection search',
                    search_ids=[f'Search {ind}' for ind, _ in enumerate(coordinates)]
                )
            ]
        )
    else:
        raise ApiError('arrival_time or departure_time should be specified')


def create_union(
    coordinates: List[Coordinates],
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
    travel_time: int,
    departure_time: Optional[datetime],
    arrival_time: Optional[datetime],
    search_range: Optional[Range]
) -> TimeMapRequest:
    if arrival_time is not None and departure_time is not None:
        raise ApiError('arrival_time and departure_time cannot be both specified')

    if arrival_time is not None:
        return TimeMapRequest(
            arrival_searches=[
                time_map.ArrivalSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    arrival_time=arrival_time,
                    transportation=transportation,
                    range=search_range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            departure_searches=[],
            unions=[
                time_map.Union(
                    id='Union search',
                    search_ids=[f'Search {ind}' for ind, _ in enumerate(coordinates)]
                )
            ],
            intersections=[]
        )
    elif departure_time is not None:
        return TimeMapRequest(
            departure_searches=[
                time_map.DepartureSearch(
                    id=f'Search {ind}',
                    coords=cur_coordinates,
                    travel_time=travel_time,
                    departure_time=departure_time,
                    transportation=transportation,
                    range=search_range
                )
                for ind, cur_coordinates in enumerate(coordinates)
            ],
            arrival_searches=[],
            unions=[
                time_map.Union(
                    id='Union search',
                    search_ids=[f'Search {ind}' for ind, _ in enumerate(coordinates)]
                )
            ],
            intersections=[]
        )
    else:
        raise ApiError('arrival_time or departure_time should be specified')


def create_routes(
    locations: List[Location],
    search_ids: Dict[str, List[str]],
    transportation: Union[PublicTransport, Driving, Ferry, Walking, Cycling, DrivingTrain],
    departure_time: Optional[datetime],
    arrival_time: Optional[datetime],
    properties: Optional[List[Property]],
    range: Optional[FullRange]
) -> RoutesRequest:
    if arrival_time is not None and departure_time is not None:
        raise ApiError('arrival_time and departure_time cannot be both specified')

    if properties is None:
        properties = [Property.TRAVEL_TIME]
    if arrival_time is not None:
        return RoutesRequest(
            locations=locations,
            arrival_searches=[
                routes.ArrivalSearch(
                    id=arrival_id,
                    arrival_location_id=arrival_id,
                    departure_location_ids=[departure_id for departure_id in departure_ids],
                    arrival_time=arrival_time,
                    transportation=transportation,
                    properties=properties,
                    range=range
                )
                for arrival_id, departure_ids in search_ids.items()
            ],
            departure_searches=[]
        )
    elif departure_time is not None:
        return RoutesRequest(
            locations=locations,
            departure_searches=[
                routes.DepartureSearch(
                    id=departure_id,
                    departure_location_id=departure_id,
                    arrival_location_ids=[arrival_id for arrival_id in arrival_ids],
                    departure_time=departure_time,
                    transportation=transportation,
                    properties=properties,
                    range=range
                )
                for departure_id, arrival_ids in search_ids.items()
            ],
            arrival_searches=[]
        )
    else:
        raise ApiError('arrival_time or departure_time should be specified')


def create_proto_request(
    origin: Coordinates,
    destinations: List[Coordinates],
    transportation: ProtoTransportation,
    travel_time: int
) -> TimeFilterFastRequest_pb2.TimeFilterFastRequest:
    request = TimeFilterFastRequest_pb2.TimeFilterFastRequest()

    request.oneToManyRequest.departureLocation.lat = origin.lat
    request.oneToManyRequest.departureLocation.lng = origin.lng

    request.oneToManyRequest.transportation.type = transportation.value.code
    request.oneToManyRequest.travelTime = travel_time
    request.oneToManyRequest.arrivalTimePeriod = TimeFilterFastRequest_pb2.TimePeriod.WEEKDAY_MORNING

    mult = math.pow(10, 5)
    for destination in destinations:
        lat_delta = round((destination.lat - origin.lat) * mult)
        lng_delta = round((destination.lng - origin.lng) * mult)
        request.oneToManyRequest.locationDeltas.extend([lat_delta, lng_delta])

    return request



from typing import List, Optional

from pydantic.main import BaseModel

from traveltimepy.dto.common import Route, Fares


class DistanceBreakdown(BaseModel):
    mode: str
    distance: int


class Property(BaseModel):
    travel_time: int
    distance: Optional[int]
    distance_breakdown: Optional[List[DistanceBreakdown]]
    fares: Optional[Fares]
    route: Optional[Route]


class Location(BaseModel):
    id: str
    properties: List[Property]


class TimeFilterResult(BaseModel):
    search_id: str
    locations: List[Location]
    unreachable: List[str]


class TimeFilterResponse(BaseModel):
    results: List[TimeFilterResult]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: TimeFilterFastResponse.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cTimeFilterFastResponse.proto\x12*com.igeolise.traveltime.rabbitmq.responses\"\xbf\x06\n\x16TimeFilterFastResponse\x12W\n\x05\x65rror\x18\x01 \x01(\x0b\x32H.com.igeolise.traveltime.rabbitmq.responses.TimeFilterFastResponse.Error\x12\x61\n\nproperties\x18\x02 \x01(\x0b\x32M.com.igeolise.traveltime.rabbitmq.responses.TimeFilterFastResponse.Properties\x1aJ\n\nProperties\x12\x13\n\x0btravelTimes\x18\x01 \x03(\x11\x12\x14\n\x0cmonthlyFares\x18\x02 \x03(\x05\x12\x11\n\tdistances\x18\x03 \x03(\x05\x1a\x63\n\x05\x45rror\x12Z\n\x04type\x18\x01 \x01(\x0e\x32L.com.igeolise.traveltime.rabbitmq.responses.TimeFilterFastResponse.ErrorType\"\xb7\x03\n\tErrorType\x12\x0b\n\x07UNKNOWN\x10\x00\x12 \n\x1cONE_TO_MANY_MUST_NOT_BE_NULL\x10\x01\x12\x1b\n\x17SOURCE_MUST_NOT_BE_NULL\x10\x02\x12#\n\x1fTRANSPORTATION_MUST_NOT_BE_NULL\x10\x03\x12\x1a\n\x16SOURCE_NOT_IN_GEOMETRY\x10\x04\x12$\n UNRECOGNIZED_TRANSPORTATION_MODE\x10\x05\x12\x1d\n\x19TRAVEL_TIME_LIMIT_TOO_LOW\x10\x06\x12\x1e\n\x1aTRAVEL_TIME_LIMIT_TOO_HIGH\x10\x07\x12\x19\n\x15\x41UTH_ERROR_NO_USER_ID\x10\x08\x12.\n*SERVICE_MISMATCH_WRONG_TRANSPORTATION_MODE\x10\t\x12\x17\n\x13SOURCE_OUT_OF_REACH\x10\n\x12/\n+INTERLEAVED_DELTAS_INVALID_COORDINATE_PAIRS\x10\x0b\x12#\n\x1f\x44ISTANCE_PROPERTY_NOT_SUPPORTED\x10\x0c\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'TimeFilterFastResponse_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TIMEFILTERFASTRESPONSE._serialized_start=77
  _TIMEFILTERFASTRESPONSE._serialized_end=908
  _TIMEFILTERFASTRESPONSE_PROPERTIES._serialized_start=291
  _TIMEFILTERFASTRESPONSE_PROPERTIES._serialized_end=365
  _TIMEFILTERFASTRESPONSE_ERROR._serialized_start=367
  _TIMEFILTERFASTRESPONSE_ERROR._serialized_end=466
  _TIMEFILTERFASTRESPONSE_ERRORTYPE._serialized_start=469
  _TIMEFILTERFASTRESPONSE_ERRORTYPE._serialized_end=908
# @@protoc_insertion_point(module_scope)



from traveltimepy import Location, Coordinates


def test_supported_locations(sdk):
    locations = [
        Location(id='Kaunas', coords=Coordinates(lat=54.900008, lng=23.957734)),
        Location(id='London', coords=Coordinates(lat=51.506756, lng=-0.12805)),
        Location(id='Bangkok', coords=Coordinates(lat=13.761866, lng=100.544818)),
        Location(id='Lisbon', coords=Coordinates(lat=38.721869, lng=-9.138549)),
        Location(id='Unsupported', coords=Coordinates(lat=68.721869, lng=-9.138549))
    ]
    response = sdk.supported_locations(locations)
    assert len(response.locations) == 4
    assert len(response.unsupported_locations) == 1



from enum import Enum


class AcceptType(Enum):
    JSON = 'application/json'
    BOUNDING_BOXES_JSON = 'application/vnd.bounding-boxes+json'
    GEO_JSON = 'application/geo+json'
    OCTET_STREAM = 'application/octet-stream'



import asyncio
from dataclasses import dataclass
from typing import TypeVar, Type, Dict

import aiohttp
from aiohttp import ClientSession, ClientResponse
from pydantic.tools import parse_raw_as
from traveltimepy.dto.requests.request import TravelTimeRequest

from traveltimepy.dto.responses.error import ResponseError
from traveltimepy.errors import ApiError
from aiohttp_retry import RetryClient, ExponentialRetry

T = TypeVar('T')
R = TypeVar('R')


@dataclass
class SdkParams:
    host: str
    limit_per_host: int


async def send_post_request_async(
    client: RetryClient,
    response_class: Type[T],
    url: str,
    headers: Dict[str, str],
    request: TravelTimeRequest
) -> T:
    async with client.post(url=url, headers=headers, data=request.json()) as resp:
        return await __process_response(response_class, resp)


async def send_post_async(
    response_class: Type[T],
    path: str,
    headers: Dict[str, str],
    request: TravelTimeRequest,
    sdk_params: SdkParams
) -> T:
    connector = aiohttp.TCPConnector(ssl=False, limit_per_host=sdk_params.limit_per_host)
    async with ClientSession(connector=connector, timeout=aiohttp.ClientTimeout(total=60 * 60 * 30)) as session:
        client = RetryClient(client_session=session, retry_options=ExponentialRetry(attempts=3))
        tasks = [
            send_post_request_async(client, response_class, f'https://{sdk_params.host}/v4/{path}', headers, part)
            for part in request.split_searches()
        ]
        responses = await asyncio.gather(*tasks)
        await client.close()
        return request.merge(responses)


def send_post(
    response_class: Type[T],
    path: str,
    headers: Dict[str, str],
    request: TravelTimeRequest,
    sdk_params: SdkParams
) -> T:
    return asyncio.run(send_post_async(response_class, path, headers, request, sdk_params))


async def send_get_async(
    response_class: Type[T],
    path: str,
    headers: Dict[str, str],
    sdk_params: SdkParams,
    params: Dict[str, str] = None
) -> T:
    connector = aiohttp.TCPConnector(ssl=False)
    async with aiohttp.ClientSession(connector=connector) as session:
        async with session.get(url=f'https://{sdk_params.host}/v4/{path}', headers=headers, params=params) as resp:
            return await __process_response(response_class, resp)


def send_get(
    response_class: Type[T],
    path: str,
    headers: Dict[str, str],
    sdk_params: SdkParams,
    params: Dict[str, str] = None
) -> T:
    return asyncio.run(send_get_async(response_class, path, headers, sdk_params, params))


async def __process_response(response_class: Type[T], response: ClientResponse) -> T:
    text = await response.text()
    if response.status != 200:
        parsed = parse_raw_as(ResponseError, text)
        msg = 'Travel Time API request failed \n{}\nError code: {}\nAdditional info: {}\n<{}>\n'.format(
            parsed.description,
            parsed.error_code,
            parsed.additional_info,
            parsed.documentation_link
        )

        raise ApiError(msg)
    else:
        return parse_raw_as(response_class, text)




def test_geocoding_search(sdk):
    response = sdk.geocoding(query='Parliament square', limit=30, within_countries=['gb', 'de'])
    assert len(response.features) > 0
    assert len(response.features) < 31


def test_geocoding_reverse(sdk):
    response = sdk.geocoding_reverse(lat=51.507281, lng=-0.132120)
    assert len(response.features) > 0
